import dataclasses
import typing


T = typing.TypeVar("T")


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class ContextRegistry:
    context: dict[type[typing.Any], typing.Any]

    def find_context(self, context_type: type[T]) -> T | None:
        if context_type and (context := self.context.get(context_type)):
            return typing.cast(T, context)

        return None



import dataclasses
import typing
import warnings

from modern_di import types
from modern_di.providers import CacheSettings, Factory


@dataclasses.dataclass(kw_only=True, slots=True)
class CacheItem:
    settings: CacheSettings[typing.Any] | None
    cache: typing.Any | None = None
    kwargs: dict[str, typing.Any] | None = None

    def _clear(self) -> None:
        if self.settings and self.settings.clear_cache:
            self.cache = None

        self.kwargs = None

    async def close_async(self) -> None:
        if self.cache and self.settings and self.settings.finalizer:
            if self.settings.is_async_finalizer:
                await self.settings.finalizer(self.cache)  # type: ignore[misc]
            else:
                self.settings.finalizer(self.cache)

        self._clear()

    def close_sync(self) -> None:
        if self.cache and self.settings and self.settings.finalizer:
            if self.settings.is_async_finalizer:
                warnings.warn(
                    f"Calling `close_sync` for async finalizer, type={type(self.cache)}",
                    RuntimeWarning,
                    stacklevel=2,
                )
                return
            self.settings.finalizer(self.cache)

        self._clear()


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class CacheRegistry:
    _items: dict[str, CacheItem] = dataclasses.field(init=False, default_factory=dict)

    def fetch_cache_item(self, provider: Factory[types.T_co]) -> CacheItem:
        return self._items.setdefault(provider.provider_id, CacheItem(settings=provider.cache_settings))

    def clear_kwargs(self) -> None:
        for cache_item in self._items.values():
            cache_item.kwargs = None

    async def close_async(self) -> None:
        for cache_item in self._items.values():
            await cache_item.close_async()

    def close_sync(self) -> None:
        for cache_item in self._items.values():
            cache_item.close_sync()



import typing

from modern_di import types
from modern_di.providers import AbstractProvider
from modern_di.scope import Scope


if typing.TYPE_CHECKING:
    from modern_di import Container


class ContextProvider(AbstractProvider[types.T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_context_type"]

    def __init__(self, *, scope: Scope = Scope.APP, context_type: type[types.T_co]) -> None:
        super().__init__(scope=scope, bound_type=context_type)
        self._context_type = context_type

    def resolve(self, container: "Container") -> types.T_co | None:
        container = container.find_container(self.scope)
        return container.context_registry.find_context(self._context_type)



import collections
import typing
import warnings

from modern_di import types
from modern_di.providers.abstract import AbstractProvider


class ProvidersRegistry:
    __slots__ = (
        "_by_name",
        "_by_type",
        "_overrides_by_name",
        "_overrides_by_type",
        "_providers_by_name",
        "_providers_by_type",
    )

    def __init__(self) -> None:
        self._providers_by_name: dict[str, AbstractProvider[typing.Any]] = {}
        self._providers_by_type: dict[type, AbstractProvider[typing.Any]] = {}
        self._overrides_by_name: dict[str, AbstractProvider[typing.Any]] = {}
        self._overrides_by_type: dict[type, AbstractProvider[typing.Any]] = {}

        self._by_name = collections.ChainMap(self._overrides_by_name, self._providers_by_name)
        self._by_type = collections.ChainMap(self._overrides_by_type, self._providers_by_type)

    def find_provider(
        self, dependency_name: str | None = None, dependency_type: type[types.T_co] | None = None
    ) -> AbstractProvider[types.T_co] | None:
        if dependency_name and (provider := self._by_name.get(dependency_name)):
            return provider

        if dependency_type and (provider := self._by_type.get(dependency_type)):
            return provider

        return None

    def add_providers(self, **kwargs: AbstractProvider[typing.Any]) -> None:
        for provider_name, provider in kwargs.items():
            if provider_name in self._providers_by_name:
                warnings.warn(
                    f"Provider is duplicated by name {provider_name}",
                    RuntimeWarning,
                    stacklevel=2,
                )
            self._providers_by_name[provider_name] = provider

            provider_type = provider.bound_type
            if not provider_type:
                continue

            if provider_type in self._providers_by_type:
                warnings.warn(
                    f"Provider is duplicated by type {provider_type}",
                    RuntimeWarning,
                    stacklevel=2,
                )

            self._providers_by_type[provider_type] = provider

    def override_provider(
        self,
        *,
        dependency_name: str | None = None,
        dependency_type: type[types.T_co] | None = None,
        new_provider: AbstractProvider[typing.Any],
    ) -> None:
        if dependency_name:
            self._overrides_by_name[dependency_name] = new_provider
        if dependency_type:
            self._overrides_by_type[dependency_type] = new_provider

    def reset_override(
        self, *, dependency_name: str | None = None, dependency_type: type[types.T_co] | None = None
    ) -> None:
        if dependency_name:
            self._overrides_by_name.pop(dependency_name, None)
        if dependency_type:
            self._overrides_by_type.pop(dependency_type, None)
        if not dependency_name and not dependency_type:
            self._overrides_by_name.clear()
            self._overrides_by_type.clear()



import typing

from modern_di import types
from modern_di.providers.abstract import AbstractProvider
from modern_di.scope import Scope


if typing.TYPE_CHECKING:
    from modern_di import Container, types


class Object(AbstractProvider[types.T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_object"]

    def __init__(
        self,
        *,
        scope: Scope = Scope.APP,
        obj: types.T_co,
        bound_type: type | None = types.UNSET,  # type: ignore[assignment]
    ) -> None:
        super().__init__(scope=scope, bound_type=bound_type if bound_type != types.UNSET else type(obj))
        self._object = obj

    def resolve(self, _: "Container") -> types.T_co:
        return self._object



import threading
import typing

import typing_extensions

from modern_di import types
from modern_di.group import Group
from modern_di.providers import Object
from modern_di.providers.abstract import AbstractProvider
from modern_di.registries.cache_registry import CacheRegistry
from modern_di.registries.context_registry import ContextRegistry
from modern_di.registries.providers_registry import ProvidersRegistry
from modern_di.scope import Scope


T_co = typing.TypeVar("T_co", covariant=True)


class Container:
    __slots__ = (
        "cache_registry",
        "context_registry",
        "lock",
        "parent_container",
        "providers_registry",
        "scope",
    )

    def __init__(
        self,
        scope: Scope = Scope.APP,
        parent_container: typing.Optional["typing_extensions.Self"] = None,
        context: dict[type[typing.Any], typing.Any] | None = None,
        groups: list[type[Group]] | None = None,
        use_lock: bool = True,
    ) -> None:
        self.lock = threading.Lock() if use_lock else None
        self.scope = scope
        self.parent_container = parent_container
        self.cache_registry = CacheRegistry()
        self.context_registry = ContextRegistry(context=context or {})
        self.providers_registry: ProvidersRegistry
        if parent_container:
            self.providers_registry = parent_container.providers_registry
        else:
            self.providers_registry = ProvidersRegistry()
            self.providers_registry.add_providers(di_container=ContainerProvider())
        if groups:
            for one_group in groups:
                self.providers_registry.add_providers(**one_group.get_providers())

    def build_child_container(
        self, context: dict[type[typing.Any], typing.Any] | None = None, scope: Scope | None = None
    ) -> "typing_extensions.Self":
        if scope and scope <= self.scope:
            msg = "Scope of child container must be more than current scope"
            raise RuntimeError(msg)

        if not scope:
            try:
                scope = self.scope.__class__(self.scope.value + 1)
            except ValueError as exc:
                msg = f"Max scope is reached, {self.scope.name}"
                raise RuntimeError(msg) from exc

        return self.__class__(scope=scope, parent_container=self, context=context)

    def find_container(self, scope: Scope) -> "typing_extensions.Self":
        container = self
        if container.scope < scope:
            msg = f"Scope {scope.name} is not initialized"
            raise RuntimeError(msg)

        while container.scope > scope and container.parent_container:
            container = container.parent_container

        if container.scope != scope:
            msg = f"Scope {scope.name} is skipped"
            raise RuntimeError(msg)

        return container

    def resolve(self, dependency_type: type[T_co] | None = None, *, dependency_name: str | None = None) -> T_co:
        provider = self.providers_registry.find_provider(
            dependency_type=dependency_type, dependency_name=dependency_name
        )
        if not provider:
            msg = f"Provider is not found, {dependency_type=}, {dependency_name=}"
            raise RuntimeError(msg)

        return typing.cast(T_co, provider.resolve(self))

    def resolve_provider(self, provider: "AbstractProvider[T_co]") -> T_co:
        return typing.cast(T_co, provider.resolve(self))

    async def close_async(self) -> None:
        if not self.parent_container:
            self.providers_registry.reset_override()
        await self.cache_registry.close_async()

    def close_sync(self) -> None:
        if not self.parent_container:
            self.providers_registry.reset_override()
        self.cache_registry.close_sync()

    def override(
        self,
        *,
        dependency_name: str | None = None,
        dependency_type: type[types.T_co] | None = None,
        mock: typing.Any,  # noqa: ANN401
    ) -> None:
        self.cache_registry.clear_kwargs()
        new_provider = mock if isinstance(mock, AbstractProvider) else Object(obj=mock)
        return self.providers_registry.override_provider(
            dependency_name=dependency_name, dependency_type=dependency_type, new_provider=new_provider
        )

    def reset_override(
        self, *, dependency_name: str | None = None, dependency_type: type[types.T_co] | None = None
    ) -> None:
        self.cache_registry.clear_kwargs()
        return self.providers_registry.reset_override(dependency_name=dependency_name, dependency_type=dependency_type)

    def __deepcopy__(self, *_: object, **__: object) -> "typing_extensions.Self":
        """Prevent cloning object."""
        return self

    def __copy__(self, *_: object, **__: object) -> "typing_extensions.Self":
        """Prevent cloning object."""
        return self


class ContainerProvider(AbstractProvider[typing.Any]):
    __slots__ = AbstractProvider.BASE_SLOTS

    def __init__(self) -> None:
        super().__init__(scope=Scope.APP, bound_type=Container)

    def resolve(self, container: Container) -> Container:
        return container



import dataclasses
import inspect
import types
import typing

from modern_di.types import UNSET


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class SignatureItem:
    arg_type: type | None = None
    args: list[type] = dataclasses.field(default_factory=list)
    is_nullable: bool = False
    default: object = UNSET

    @classmethod
    def from_type(cls, type_: type, default: object = UNSET) -> "SignatureItem":
        result: dict[str, typing.Any] = {"default": default}
        if isinstance(type_, types.GenericAlias):
            result["arg_type"] = type_.__origin__
            result["args"] = list(type_.__args__)
        elif isinstance(type_, (types.UnionType, typing._UnionGenericAlias)):  # type: ignore[attr-defined]  # noqa: SLF001
            args = [x.__origin__ if isinstance(x, types.GenericAlias) else x for x in type_.__args__]
            if types.NoneType in args:
                result["is_nullable"] = True
                args.remove(types.NoneType)
            if len(args) > 1:
                result["args"] = args
            elif args:
                result["arg_type"] = args[0]
        elif isinstance(type_, type):
            result["arg_type"] = type_
        return cls(**result)


def parse_creator(creator: typing.Callable[..., typing.Any]) -> tuple[SignatureItem, dict[str, SignatureItem]]:
    try:
        sig = inspect.signature(creator)
    except ValueError:
        return SignatureItem.from_type(typing.cast(type, creator)), {}

    param_hints = {}
    for param_name, param in sig.parameters.items():
        if param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
            continue
        default = UNSET
        if param.default is not param.empty:
            default = param.default
        if param.annotation is not param.empty:
            param_hints[param_name] = SignatureItem.from_type(param.annotation, default=default)
        else:
            param_hints[param_name] = SignatureItem(default=default)
    if sig.return_annotation:
        return_sig = SignatureItem.from_type(sig.return_annotation)
    elif isinstance(creator, type):
        return_sig = SignatureItem.from_type(creator)
    else:
        return_sig = SignatureItem()

    return return_sig, param_hints
